from __future__ import division

import httplib2, base64, json
import logging, datetime, os, platform

import paypalrestsdk.util as util
from paypalrestsdk.exceptions import *
from paypalrestsdk.version    import __version__

class Api:

  # User-Agent for HTTP request
  library_details = "httplib2 %s; python %s"%(httplib2.__version__, platform.python_version())
  user_agent = "PayPalSDK/rest-sdk-python %s (%s)"%(__version__, library_details)

  # Create API object
  # == Example
  #   import paypalrestsdk
  #   api = paypalrestsdk.Api( mode="sandbox",
  #          client_id='CLIENT_ID', client_secret='CLIENT_SECRET', ssl_options={} )
  def __init__(self, options = {}, **args):
    args = util.merge_dict(options, args)

    self.mode           = args.get("mode", "sandbox")
    self.endpoint       = args.get("endpoint", self.default_endpoint())
    self.token_endpoint = args.get("token_endpoint", self.endpoint)
    self.client_id      = args.get("client_id")
    self.client_secret  = args.get("client_secret")
    self.ssl_options    = args.get("ssl_options", {})

    self.token_hash       = None
    self.token_request_at = None
    if args.get("token"):
      self.token_hash     = { "access_token": args.get("token"), "token_type": "Bearer" }

    self.options = args

  # Default endpoint
  def default_endpoint(self):
    if self.mode == "live" :
      return "https://api.paypal.com"
    else:
      return "https://api.sandbox.paypal.com"

  # Find basic auth
  def basic_auth(self):
    credentials = "%s:%s"%(self.client_id, self.client_secret)
    return base64.b64encode(credentials.encode('utf-8')).decode('utf-8').replace("\n", "")

  # Generate token_hash
  def get_token_hash(self):
    self.validate_token_hash()
    if self.token_hash == None :
      self.token_request_at = datetime.datetime.now()
      self.token_hash = self.http_call(util.join_url(self.token_endpoint, "/v1/oauth2/token"), "POST",
        body = "grant_type=client_credentials",
        headers = { "Authorization": ("Basic %s" % self.basic_auth()),
          "Content-Type": "application/x-www-form-urlencoded",
          "Accept": "application/json", "User-Agent": self.user_agent } )
    return self.token_hash

  # Validate expires_in
  def validate_token_hash(self):
    if self.token_request_at and self.token_hash and self.token_hash.get("expires_in") != None :
      delta = datetime.datetime.now() - self.token_request_at
      duration = (delta.microseconds + (delta.seconds + delta.days * 24 * 3600) * 10**6) / 10**6
      if duration > self.token_hash.get("expires_in"):
        self.token_hash = None

  # Get access token
  def get_token(self):
    return self.get_token_hash()["access_token"]

  # Get token type
  def get_token_type(self):
    return self.get_token_hash()["token_type"]

  # Make HTTP call and Format Response
  # == Example
  #   api.request("https://api.sandbox.paypal.com/v1/payments/payment?count=10", "GET", {})
  #   api.request("https://api.sandbox.paypal.com/v1/payments/payment", "POST", "{}", {} )
  def request(self, url, method, body = None, headers = {}):

    http_headers = util.merge_dict(self.headers(), headers)

    if http_headers.get('PayPal-Request-Id'):
      logging.info('PayPal-Request-Id: %s'%(http_headers['PayPal-Request-Id']))

    try:
      return self.http_call(url, method, body= body, headers= http_headers)

    # Format Error message for bad request
    except BadRequest as error:
      return { "error": json.loads(error.content) }

    # Handle Exipre token
    except UnauthorizedAccess as error:
      if(self.token_hash and self.client_id):
        self.token_hash = None
        return self.request(url, method, body, headers)
      else:
        raise error

  # Make http Call
  def http_call(self, url, method, **args):
    logging.info('Request[%s]: %s'%(method, url))
    http = httplib2.Http(**self.ssl_options)
    start_time = datetime.datetime.now()
    response, content = http.request(url, method, **args)
    duration   = datetime.datetime.now() - start_time
    logging.info('Response[%d]: %s, Duration: %s.%ss'%(response.status, response.reason, duration.seconds, duration.microseconds))
    return self.handle_response(response, content.decode('utf-8'))

  # Validate HTTP response
  def handle_response(self, response, content):
    status = response.status
    if status in [ 301, 302, 303, 307 ] :
      raise Redirection(response, content)
    elif status >= 200 and status <= 299 :
      if content :
        return json.loads(content)
      else:
        return {}
    elif status == 400 :
      raise BadRequest(response, content)
    elif status == 401 :
      raise UnauthorizedAccess(response, content)
    elif status == 403 :
      raise ForbiddenAccess(response, content)
    elif status == 404 :
      raise ResourceNotFound(response, content)
    elif status == 405 :
      raise MethodNotAllowed(response, content)
    elif status == 409 :
      raise ResourceConflict(response, content)
    elif status == 410 :
      raise ResourceGone(response, content)
    elif status == 422 :
      raise ResourceInvalid(response, content)
    elif status >= 401 and status <= 499 :
      raise ClientError(response, content)
    elif status >= 500 and status <= 599 :
      raise ServerError(response, content)
    else:
      raise ConnectionError(response, content, "Unknown response code: #{response.code}")

  # Default HTTP headers
  def headers(self):
    return { "Authorization": ("%s %s" %(self.get_token_type(), self.get_token())),
        "Content-Type": "application/json",
        "Accept": "application/json",
        "User-Agent": self.user_agent }

  # Make GET request
  # == Example
  #   api.get("v1/payments/payment?count=1")
  #   api.get("v1/payments/payment/PAY-1234")
  def get(self, action, headers = {}):
    return self.request(util.join_url(self.endpoint, action), 'GET', headers = headers)

  # Make POST request
  # == Example
  #   api.post("v1/payments/payment", { 'indent': 'sale' })
  #   api.post("v1/payments/payment/PAY-1234/execute", { 'payer_id': '1234' })
  def post(self, action, params = {}, headers = {}):
    return self.request(util.join_url(self.endpoint, action), 'POST', body= json.dumps(params), headers = headers )

  # Make DELETE request
  def delete(self, action, headers = {}):
    return self.request(util.join_url(self.endpoint, action), 'DELETE', headers = headers)


global __api__
__api__ = None

# Get default api
def default():
  global __api__
  if __api__ is None :
    try:
      client_id     = os.environ["PAYPAL_CLIENT_ID"]
      client_secret = os.environ["PAYPAL_CLIENT_SECRET"]
    except KeyError as error :
      raise MissingConfig("Required PAYPAL_CLIENT_ID and PAYPAL_CLIENT_SECRET. Refer https://github.com/paypal/rest-api-sdk-python#configuration")

    __api__ = Api( mode= os.environ.get("PAYPAL_MODE", "sandbox"),
      client_id= client_id, client_secret= client_secret)
  return __api__

# Create new default api object with given configuration
def set_config(options = {}, **config):
  global __api__
  __api__ = Api(options, **config)
  return __api__

configure = set_config



from paypalrestsdk.resource import List, Find, Create, Post

# == Example
#  payment_histroy = Payment.all({"count": 5})
#  payment = Payment.find("PAY-1234")
#  payment = Payment.new({"intent": "sale"})
#
#  payment.create()     # return True or False
#  payment.execute({"payer_id": 1234})  # return True or False
class Payment(List, Find, Create, Post):

  path = "v1/payments/payment"

  def execute(self, attributes):
    return self.post('execute', attributes, self)

Payment.convert_resources['payments'] = Payment
Payment.convert_resources['payment']  = Payment

# == Example
#  sale = Sale.find("98765432")
#
#  refund = sale.refund({"amount": {"total": "1.00", "currency": "USD"}})
#  refund.success()   # return True or False
class Sale(Find, Post):

  path = "v1/payments/sale"

  def refund(self, attributes):
    return self.post('refund', attributes, Refund)

Sale.convert_resources['sales'] = Sale
Sale.convert_resources['sale']  = Sale


# == Example
#  refund = Refund.find("12345678")
class Refund(Find):

  path = "v1/payments/refund"

Refund.convert_resources['refund'] = Refund

# == Example
#   authorization = Authorization.find("")
#   capture = authorization.capture({ "amount": { "currency": "USD", "total": "1.00" } })
#   authorization.void() # return True or False
class Authorization(Find, Post):

  path = "v1/payments/authorization"

  def capture(self, attributes):
    return self.post('capture', attributes, Capture)

  def void(self):
    return self.post('void', {}, self)

Authorization.convert_resources['authorization'] = Authorization

# == Example
#   capture = Capture.find("")
#   refund = capture.refund({ "amount": { "currency": "USD", "total": "1.00" })
class Capture(Find, Post):

  path = "v1/payments/capture"

  def refund(self, attributes):
    return self.post('refund', attributes, Refund)

Capture.convert_resources['capture'] = Capture



from paypalrestsdk.resource import Find, Create, Delete

# == Example
#   credit_card = CreditCard.find("CARD-5BT058015C739554AKE2GCEI")
#   credit_card = CreditCard.new({'type': 'visa'})
#
#   credit_card.create()  # return True or False
class CreditCard(Find, Create, Delete):

  path = "v1/vault/credit-card"

CreditCard.convert_resources['credit_card'] = CreditCard



from paypalrestsdk.resource import Resource
import paypalrestsdk.util as util
import paypalrestsdk.api  as api
from paypalrestsdk.version    import __version__

class Base(Resource):

  user_agent = "PayPalSDK/openid-connect-python %s (%s)"%(__version__, api.Api.library_details)

  @classmethod
  def post(klass, action, options = {}, headers = {}):
    url  = util.join_url(endpoint(), action)
    body = util.urlencode(options)
    headers = util.merge_dict({
      'User-Agent': klass.user_agent,
      'Content-Type': 'application/x-www-form-urlencoded'}, headers)
    data = api.default().http_call(url, 'POST', body= body, headers= headers)
    return klass(data)

class Tokeninfo(Base):

  path = "v1/identity/openidconnect/tokenservice"

  @classmethod
  def create(klass, options = {}):
    if isinstance(options, str):
      options = { 'code': options }

    options = util.merge_dict({
      'grant_type': 'authorization_code',
      'client_id': client_id(),
      'client_secret': client_secret() }, options)

    return klass.post(klass.path, options)

  @classmethod
  def create_with_refresh_token(klass, options = {}):
    if isinstance(options, str):
      options = { 'refresh_token': options }

    options = util.merge_dict({
      'grant_type': 'refresh_token',
      'client_id': client_id(),
      'client_secret': client_secret() }, options)

    return klass.post(klass.path, options)

  @classmethod
  def authorize_url(klass, options = {}):
    return authorize_url(options)

  def logout_url(self, options = {}):
    options = util.merge_dict({
      'id_token': self.id_token }, options)
    return logout_url(options)

  def refresh(self, options = {}):
    options = util.merge_dict({
      'refresh_token': self.refresh_token }, options)
    tokeninfo = self.__class__.create_with_refresh_token(options)
    self.merge(tokeninfo.to_dict())
    return self

  def userinfo(self, options = {}):
    options = util.merge_dict({
      'access_token': self.access_token }, options)
    return Userinfo.get(options)

class Userinfo(Base):

  path = "v1/identity/openidconnect/userinfo"

  @classmethod
  def get(klass, options = {}):

    if isinstance(options, str):
      options = { 'access_token': options }
    options = util.merge_dict({ 'schema': 'openid' }, options)

    return klass.post(klass.path, options)

def endpoint():
  return api.default().options.get("openid_endpoint", "https://api.paypal.com/")

def client_id():
  return api.default().options.get("openid_client_id", api.default().client_id)

def client_secret():
  return api.default().options.get("openid_client_secret", api.default().client_secret)

def redirect_uri():
  return api.default().options.get("openid_redirect_uri")

start_session_path = "https://www.paypal.com/webapps/auth/protocol/openidconnect/v1/authorize"
end_session_path   = "https://www.paypal.com/webapps/auth/protocol/openidconnect/v1/endsession"

def authorize_url(options = {}):
  options = util.merge_dict({
    'response_type': 'code',
    'scope': 'openid',
    'client_id': client_id(),
    'redirect_uri': redirect_uri() }, options)
  return util.join_url_params(start_session_path, options)

def logout_url(options = {}):
  options = util.merge_dict({
    'logout': 'true',
    'redirect_uri': redirect_uri() }, options)
  return util.join_url_params(end_session_path, options)



import re
try:
  from urllib.parse import urlencode
except ImportError:
  from urllib import urlencode

# Join given url
# == Example
#   api.join_url("example.com", "index.html") # Return "example.com/index.html"
def join_url(url, *paths):
  for path in paths:
    url = re.sub(r'/?$', re.sub(r'^/?', '/', path), url)
  return url


def join_url_params(url, params):
  return url + "?" + urlencode(params)

# Merge dict
def merge_dict(data, *override):
  dict_list = list(data.items())
  for value in override:
    dict_list = dict_list + list(value.items())
  return dict(dict_list)



import paypalrestsdk.util as util
import paypalrestsdk.api  as api
import uuid, json

# Base class for all REST service
class Resource(object):

  convert_resources = {}

  def __init__(self, attributes = {}):
    super(Resource, self).__setattr__('__data__', {})
    super(Resource, self).__setattr__('error', None)
    super(Resource, self).__setattr__('headers', {})
    super(Resource, self).__setattr__('header', {})
    super(Resource, self).__setattr__('request_id', None)
    self.merge(attributes)

  # Generate uniq request id
  def generate_request_id(self):
    if self.request_id == None :
      self.request_id = str(uuid.uuid4())
    return self.request_id

  # Generate HTTP header
  def http_headers(self):
    return util.merge_dict(self.header, self.headers,
        { 'PayPal-Request-Id': self.generate_request_id() })

  def __str__(self):
    return self.__data__.__str__()

  def __repr__(self):
    return self.__data__.__str__()

  # Getter
  def __getattr__(self, name):
    return self.__data__.get(name)

  # Setter
  def __setattr__(self, name, value):
    try:
      # Handle attributes(error, header, request_id)
      super(Resource, self).__getattribute__(name)
      super(Resource, self).__setattr__(name, value)
    except AttributeError:
      self.__data__[name] = self.convert(name, value)

  # return True if no error
  def success(self):
    return self.error == None

  # Merge new attributes
  def merge(self, new_attributes):
    for k in new_attributes:
      self.__setattr__(k, new_attributes[k])

  # Convert the attribute values to configured class.
  def convert(self, name, value):
    if isinstance(value, dict):
      klass = self.convert_resources.get(name, Resource)
      return klass(value)
    elif isinstance(value, list):
      new_list = []
      for obj in value:
        new_list.append(self.convert(name, obj))
      return new_list
    else:
      return value

  def __getitem__(self, key):
    return self.__data__[key]

  def __setitem__(self, key, value):
    self.__data__[key] = self.convert(key, value)

  def to_dict(self):

    def parse_object(value):
      if isinstance(value, Resource):
        return value.to_dict()
      elif isinstance(value, list):
        new_list = []
        for obj in value:
          new_list.append(parse_object(obj))
        return new_list
      else:
        return value

    data = {}
    for key in self.__data__:
      data[key] = parse_object(self.__data__[key])
    return data

# == Example
#   payment = Payment.find("PAY-1234")
class Find(Resource):
  @classmethod
  def find(klass, resource_id):
    url = util.join_url(klass.path, str(resource_id))
    return klass(api.default().get(url))

# == Example
#   payment_histroy = Payment.all({'count': 2})
class List(Resource):
  list_class = Resource

  @classmethod
  def all(klass, params = None):
    if params == None:
      url = klass.path
    else:
      url = util.join_url_params(klass.path, params)
    return klass.list_class(api.default().get(url))

# == Example
#   payment = Payment({})
#   payment.create() # return True or False
class Create(Resource):
  def create(self):
    new_attributes = api.default().post(self.path, self.to_dict(), self.http_headers())
    self.error = None
    self.merge(new_attributes)
    return self.success()

# == Example
#   credit_card.delete()
class Delete(Resource):
  def delete(self):
    url = util.join_url(self.path, str(self['id']))
    new_attributes = api.default().delete(url)
    self.error = None
    self.merge(new_attributes)
    return self.success()

# == Example
#   payment.post("execute", {'payer_id': '1234'}, payment)  # return True or False
#   sale.post("refund", {'payer_id': '1234'})  # return Refund object
class Post(Resource):
  def post(self, name, attributes = {}, klass = Resource):
    url = util.join_url(self.path, str(self['id']), name)
    if not isinstance(attributes, Resource):
      attributes = Resource(attributes)
    new_attributes = api.default().post(url, attributes.to_dict(), attributes.http_headers())
    if isinstance(klass, Resource):
      klass.error = None
      klass.merge(new_attributes)
      return self.success()
    else:
      return klass(new_attributes)




class ConnectionError(Exception):
  def __init__(self, response, content = None, message = None):
    self.response = response
    self.content  = content
    self.message  = message

  def __str__(self):
    message = "Failed."
    if hasattr(self.response, 'status'):
      message = message + "  Response status = %s."%(self.response.status)
    if hasattr(self.response, 'reason'):
      message = message + "  Response message = %s."%(self.response.reason)
    return message

# Raised when a Timeout::Error occurs.
class TimeoutError(ConnectionError):
  def __init__(self, message):
    self.message = message

  def __str__(self):
    return self.message

# Raised when a OpenSSL::SSL::SSLError occurs.
class SSLError(ConnectionError):
  def __init__(self, message):
    self.message = message

  def __str__(self):
    return self.message

# 3xx Redirection
class Redirection(ConnectionError):
  def __str__(self):
    message = super(Redirection, self).__str__()
    if self.response.get('Location'):
      message = "%s => %s"%(message, self.response.get('Location'))
    return message

class MissingParam(TypeError): pass

class MissingConfig(Exception): pass

# 4xx Client Error
class ClientError(ConnectionError): pass

# 400 Bad Request
class BadRequest(ClientError): pass

# 401 Unauthorized
class UnauthorizedAccess(ClientError): pass

# 403 Forbidden
class ForbiddenAccess(ClientError): pass

# 404 Not Found
class ResourceNotFound(ClientError): pass

# 409 Conflict
class ResourceConflict(ClientError): pass

# 410 Gone
class ResourceGone(ClientError): pass

# 5xx Server Error
class ServerError(ConnectionError): pass

# 405 Method Not Allowed
class MethodNotAllowed(ClientError):
  def allowed_methods(self):
    return self.response['Allow']



from paypalrestsdk.api            import Api, set_config, configure
from paypalrestsdk.payments       import Payment, Sale, Refund, Authorization, Capture
from paypalrestsdk.vault          import CreditCard
from paypalrestsdk.openid_connect import Tokeninfo, Userinfo
from paypalrestsdk.exceptions     import *
from paypalrestsdk.version        import __version__




from distutils.core import setup

import sys, os
sys.path.insert(0, os.path.join(os.path.dirname(__file__), 'paypalrestsdk'))
from version import __version__

setup(
  name='paypalrestsdk',
  version= __version__,
  author='PayPal',
  author_email='DL-PP-Platform-Python-SDK@ebay.com',
  packages=['paypalrestsdk'],
  scripts=[],
  url='https://developer.paypal.com',
  license='LICENSE.txt',
  description='The PayPal REST SDK provides Python APIs to create, process and manage payment',
  long_description="""
    The PayPal REST SDK provides Python APIs to create, process and manage payment.

    1. https://github.com/paypal/rest-api-sdk-python/ - README and Samples
    2. https://developer.paypal.com/webapps/developer/docs/api/ - API Reference
  """,
  install_requires=['httplib2'],
  classifiers=[
    'Intended Audience :: Developers',
    'Natural Language :: English',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Programming Language :: Python :: 2',
    'Programming Language :: Python :: 3',
    'Topic :: Software Development :: Libraries :: Python Modules'
  ],
  keywords="paypal rest sdk",
)
